import torch
from torch.utils.data import TensorDataset
import torchvision.datasets as datasets
from torchvision import transforms
import numpy as np
import matplotlib.pyplot as plt
import collections
from tqdm import tqdm
import pdb
from torch.utils.data import Dataset

class CELEBA(Dataset):
    def __init__(self,train=True):
        transform = transforms.Compose([
        # transforms.Scale(image_size),
        # transforms.Resize(image_size),
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
])
        split = 'train' if train else 'test'
        self.dataset = datasets.CelebA(root='../../data', split=split, download=True, transform=transform)

    def __getitem__(self,idx):
        img, attr = self.dataset.__getitem__(idx)
        size_1 = 218
        size_2 = 178
        x = [[int(i/size_2)/size_1,(i%size_2)/size_2] for i in range(size_1*size_2)]
        x = torch.from_numpy(np.array(x))
        x = x.type(torch.FloatTensor)
        
        img = img.permute([1,2,0]).contiguous()
        y = img.view(-1,3)
        random_idx = np.random.choice(size_1*size_2,size=1500,replace=False)
        random_idx = torch.from_numpy(random_idx)
        x = torch.index_select(x,0,random_idx)
        y = torch.index_select(y,0,random_idx)
        return x,y,random_idx,attr
    def __len__(self):
        return len(self.dataset)

def plot(target_x, target_y, context_x, context_y, pred_y, var, it, model_name):

    target_x, target_y, context_x, context_y, pred_y, var \
      = target_x[0].to('cpu').data, target_y[0].to('cpu').data, context_x[0].to('cpu').data, context_y[0].to('cpu').data, pred_y[0].to('cpu').data, var[0].to('cpu').data
    target_y = target_y.view([28,28])
    pred_y = pred_y.view([28,28])
    var = var.view([28,28])

    plt.imshow(target_y,cmap='gray',vmin=0,vmax=1)
    plt.savefig('images/%s%d_target.png'%(model_name,it))
    plt.close()


    plt.imshow(pred_y,cmap='gray',vmin=0,vmax=1)
    plt.savefig('images/%s%d_pred.png'%(model_name,it))
    plt.close()

    plt.imshow(var,cmap='gray',vmin=var.min(),vmax=var.max())
    plt.savefig('images/%s%d_var.png'%(model_name,it))
    plt.close()

    pdb.set_trace()
