import torch
import torch.nn as nn
import torch.nn.functional as F
from models.anp import *

import pdb
from collections import OrderedDict
import numpy as np


class Classification_Model(nn.Module):
    def __init__(self):
        super(Classification_Model, self).__init__()
        self.ConvLayer1 = nn.Sequential(
            nn.Conv2d(3, 64, 3), 
            nn.BatchNorm2d(64),
            nn.ReLU(), 
            nn.MaxPool2d(2), 
        )
        self.ConvLayer2 = nn.Sequential(
            nn.Conv2d(64, 128, 3), 
            nn.BatchNorm2d(128), 
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer3 = nn.Sequential(
            nn.Conv2d(128, 256, 3), 
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer4 = nn.Sequential(
            nn.Conv2d(256, 512, 3), 
            nn.BatchNorm2d(512),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.Linear1 = nn.Sequential(nn.Linear(512 * 11 * 9, 1024), nn.ReLU())
        self.Linear2 = nn.Sequential(nn.Linear(1024, 256), nn.ReLU())
        self.Linear3 = nn.Sequential(nn.Linear(256, 40), nn.Sigmoid())
    def forward(self,x,y):        
        x = self.ConvLayer1(x)
        x = self.ConvLayer2(x)
        x = self.ConvLayer3(x)
        x = self.ConvLayer4(x)
        x = x.view(x.shape[0],-1)

        pred = self.Linear3(self.Linear2(self.Linear1(x))).squeeze(1)

        loss = F.binary_cross_entropy(pred, y, reduction='none')

        return loss, pred

class to2D(nn.Module):
    def __init__(self):
        super(to2D, self).__init__()
    def forward(self,x):
        return x.squeeze(3).squeeze(2)

class to4D(nn.Module):
    def __init__(self):
        super(to4D, self).__init__()
    def forward(self,x):
        return x.unsqueeze(1)


class Sampler(nn.Module):
    def __init__(self, num_output_points, num_conv_layers, num_fc_layers, x_dim, y_dim, hidden_dim):
        super(Sampler, self).__init__()
        self._module = nn.Sequential(to4D(),
                                    nn.Conv2d(in_channels=1,
                                                out_channels=hidden_dim,
                                                kernel_size=[1,x_dim+y_dim]),
                                    nn.BatchNorm2d(hidden_dim),
                                    nn.ReLU(),
                                    nn.Conv2d(in_channels=hidden_dim,
                                                out_channels=hidden_dim,
                                                kernel_size=[1,1]),
                                    nn.BatchNorm2d(hidden_dim),
                                    nn.ReLU(),
                                    nn.Conv2d(in_channels=hidden_dim,
                                                out_channels=hidden_dim,
                                                kernel_size=[1,1]),
                                    nn.BatchNorm2d(hidden_dim),
                                    nn.ReLU(),
                                    nn.Conv2d(in_channels=hidden_dim,
                                                out_channels=hidden_dim,
                                                kernel_size=[1,1]),
                                    nn.BatchNorm2d(hidden_dim),
                                    nn.ReLU(),
                                    nn.AdaptiveMaxPool2d([1,1]),
                                    to2D(),
                                    nn.Linear(hidden_dim,hidden_dim),
                                    nn.BatchNorm1d(hidden_dim),
                                    nn.Linear(hidden_dim,hidden_dim),
                                    nn.BatchNorm1d(hidden_dim),
                                    nn.Linear(hidden_dim,hidden_dim),
                                    nn.BatchNorm1d(hidden_dim),
                                    nn.Linear(hidden_dim,num_output_points*(x_dim+y_dim)),
                                    nn.BatchNorm1d(num_output_points*(x_dim+y_dim))
                                    )
        self.beta = 1
        self.gamma = 1
        self.theta = 0
        self.alpha = 100
    
    def matching_loss(self,G,P):
        # G: [B,k,d]
        # P: [B,N,d]
        G_f = G.unsqueeze(2)
        G_f = G_f.repeat([1,1,P.shape[1],1])
        P_f = P.unsqueeze(1)
        # G_f: [B,k,N,d]
        # P_f: [B,1,N,d]
        dist = G_f - P_f 
        dist = torch.norm(dist,dim=3)**2
        dist_minP = dist.min(2)[0]
        
        L_f = dist_minP.mean(1)
        L_m = dist_minP.max(1)[0]

        dist_minG = dist.min(1)[0]
        L_b = dist_minG.mean(1)

        matching_loss = self.alpha*(L_f + self.beta*L_m + (self.gamma+self.theta*G.shape[1])*L_b)

        return matching_loss

    def match(self, G, P):
        # G: [B,k,d]
        # P: [B,N,d]
        
        G_f = G.unsqueeze(2)
        G_f = G_f.repeat([1,1,P.shape[1],1])
        P_f = P.unsqueeze(1)
        dist = G_f - P_f 
        dist = torch.norm(dist,dim=3)

        index = dist.min(2)[1]
        use_fps = True
        if use_fps:
            subset = []
            for b in range(index.shape[0]):
                idx = set(index[b].to('cpu').data.numpy())
                sub = P[b][list(idx)].unsqueeze(0)
                sub_com = P[b][list(set(range(len(P[b]))).difference(idx))].unsqueeze(1)
                for _ in range(index.shape[1]-len(idx)):
                    d = torch.sqrt(torch.sum((sub-sub_com)**2,-1)).sum(1)
                    max_pos = torch.max(d,0)[1].item()
                    sub = torch.cat([sub,sub_com[max_pos][None,:,:]],1)
                    sub_com = torch.cat([sub_com[:max_pos],sub_com[max_pos+1:]],0)
                subset.append(sub)
            subset = torch.cat(subset,0)
            return subset
        else:
            index = index[:,:,None].repeat([1,1,P.shape[-1]])
            return torch.gather(P,1,index)

    def forward(self, x, y, i):
        original_point_clouds = torch.cat([x,y],2)

        sub_point_clouds = self._module(original_point_clouds)
        sub_point_clouds = sub_point_clouds.view([original_point_clouds.shape[0],-1,original_point_clouds.shape[2]])
        assert i<=sub_point_clouds.shape[1], 'Check the number of output points' 
        sub_point_clouds = sub_point_clouds[:,:i,:]


        matching_loss = torch.Tensor([0.0]).to(sub_point_clouds.device)
        if self.training:
            matching_loss = self.matching_loss(sub_point_clouds,original_point_clouds)
        else:
            sub_point_clouds = self.match(sub_point_clouds,original_point_clouds)
        
        temp = torch.split(sub_point_clouds, x.shape[2], 2)
        sub_x = temp[0]
        sub_y = torch.cat(temp[1:],2)

        return sub_x, sub_y, matching_loss


class Model(nn.Module):
    def __init__(self, config):
        super(Model, self).__init__()
        self._sampler = Sampler(config.max_output_points, config.sampler_num_conv_layers, config.sampler_num_fc_layers, config.x_dim, config.y_dim, config.hidden_dim)
        self._decoder = DeterministicANP(config.CNP_encoder_num_layers, config.CNP_decoder_num_layers, config.x_dim, config.y_dim, config.hidden_dim)
        self.classify_model = Classification_Model() 

    def forward(self, x, y, index, attr, subset_size):
        subset_x, subset_y, matching_loss = self._sampler(x,y,subset_size)
        log_p , mu, var = self._decoder(subset_x, subset_y , x, y,subset_x.new_ones([subset_x.shape[0],subset_x.shape[1],1]))
        log_p_matched = None
        with torch.no_grad():
            self._sampler.eval()
            subset_x_matched, subset_y_matched, _ = self._sampler(x,y,subset_size)
            self._sampler.train()
        _ , mu_matched, var_matched = self._decoder(subset_x_matched, subset_y_matched , x, y,subset_x.new_ones([subset_x.shape[0],subset_x.shape[1],1]))
        img = y.new_zeros([y.shape[0],218*178,y.shape[2]])
        img.scatter_(1,index,mu_matched)
        img = img.view(-1,218,178,img.shape[-1]).permute(0,3,1,2).contiguous()
        classify_loss, pred = self.classify_model(img,attr)

        
        return subset_x, subset_y , matching_loss, log_p, mu, var, classify_loss, pred
class DataParallel_Model(nn.Module):
    def __init__(self, config):
        super(DataParallel_Model, self).__init__()
        self._model = Model(config)
        self._model = nn.DataParallel(self._model)
        self.num_points_jump = config.num_points_jump
        self.max_output_points = config.max_output_points

    def forward(self, x, y, index, attr, subset_size=500):
        if self.training:
            i = np.random.choice(self.max_output_points)
            subset_size = int(i/self.num_points_jump)*self.num_points_jump+self.num_points_jump
        else:
            subset_size = subset_size

        subset_x, subset_y , matching_loss, log_p, mu, var, classify_loss, pred = self._model(x,y,subset_size)

        loss_p = -log_p.sum(1).mean()
        matching_loss = matching_loss.mean()
        loss = loss_p + matching_loss

        train_dict = OrderedDict()
        train_dict['Reconstruction loss'] = loss_p.to('cpu').item()
        train_dict['Matching loss'] = matching_loss.to('cpu').item()
        
        classify_loss = classify_loss.sum(1).mean()
        loss += classify_loss
        pred_dict = OrderedDict()
        if pred_greedy is not None:
            pred_dict['Main'] = pred.to('cpu').data


        test_dict = OrderedDict()
        test_dict['subset_x'] = subset_x
        test_dict['subset_y'] = subset_y
        test_dict['mu'] = mu 
        test_dict['var'] = var

        return loss, train_dict, test_dict
