import torch
import torch.nn as nn
import pdb
import torch.nn.functional as F
from torch.distributions.relaxed_bernoulli import RelaxedBernoulli
from torch.distributions.bernoulli import Bernoulli
import numpy as np
from collections import OrderedDict
from models.anp import *
from torch.distributions.relaxed_categorical import RelaxedOneHotCategorical
from torch.distributions.one_hot_categorical import OneHotCategorical
from models.transformer_modules import *

class Classification_Model(nn.Module):
    def __init__(self):
        super(Classification_Model, self).__init__()
        self.ConvLayer1 = nn.Sequential(
            nn.Conv2d(3, 64, 3), 
            nn.BatchNorm2d(64),
            nn.ReLU(), 
            nn.MaxPool2d(2), 
        )
        self.ConvLayer2 = nn.Sequential(
            nn.Conv2d(64, 128, 3), 
            nn.BatchNorm2d(128), 
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer3 = nn.Sequential(
            nn.Conv2d(128, 256, 3), 
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer4 = nn.Sequential(
            nn.Conv2d(256, 512, 3), 
            nn.BatchNorm2d(512),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.Linear1 = nn.Sequential(nn.Linear(512 * 11 * 9, 1024), nn.ReLU())
        self.Linear2 = nn.Sequential(nn.Linear(1024, 256), nn.ReLU())
        self.Linear3 = nn.Sequential(nn.Linear(256, 40), nn.Sigmoid())
    def forward(self,x,y):        
        x = self.ConvLayer1(x)
        x = self.ConvLayer2(x)
        x = self.ConvLayer3(x)
        x = self.ConvLayer4(x)
        x = x.view(x.shape[0],-1)

        pred = self.Linear3(self.Linear2(self.Linear1(x))).squeeze(1)

        loss = F.binary_cross_entropy(pred, y, reduction='none')

        return loss, pred

#### Encoder: Subset Selection 
class BerDrop(nn.Module):
    def __init__(self, num_gates,reg_scale=1,
            alpha=0.7, thres=0.499, construct_real_mask=True):
        super(BerDrop, self).__init__()
        self.num_gates = num_gates
        self.reg_scale = reg_scale
        self.alpha = alpha
        self.thres = thres
        self.construct_real_mask = construct_real_mask

        self.q_network = nn.Sequential(
                                      nn.BatchNorm1d(256),
                                      nn.Linear(in_features=256, out_features=128, bias=False),
                                      nn.BatchNorm1d(128),
                                      nn.ReLU(),
                                      nn.Linear(in_features=128, out_features=128, bias=False),
                                      nn.BatchNorm1d(128),
                                      nn.ReLU(),
                                      nn.Linear(in_features=128, out_features=1, bias=False),
                                      nn.BatchNorm1d(1)
                                      #nn.Linear(in_features=10, out_features=1, bias=True)
                                      )

        self.q_network.apply(self.init_weights)

    @staticmethod
    def init_weights(m):
        if type(m) == nn.Linear:
            nn.init.xavier_uniform_(m.weight)
            #m.bias.data.fill_(0.5)

    def forward(self, x, get_reg=True):
        batch_size = x.shape[0]
        set_size = x.shape[1]

        x = x.view(batch_size*set_size,-1)
        p = torch.sigmoid(self.q_network(x))
        p = p.view(batch_size,set_size,self.num_gates)
        p = 1e-10 + (1-2e-10)*p

        if self.training:
            temp = torch.Tensor([0.05])
            if torch.cuda.is_available():
                temp = temp.cuda()
            p_z = RelaxedBernoulli(temp, probs=p)
            z = p_z.rsample()
        else:
            z = Bernoulli(probs=p).sample()

        reg = None
        if get_reg:
            reg = p*torch.log(p/self.alpha+1e-10) + (1-p)*torch.log((1-p)/(1-self.alpha)+1e-10)

        return z, reg*self.reg_scale

class AutoRegressive(nn.Module):
    def __init__(self, num_layers, in_dim, hidden_dim):
        super(AutoRegressive, self).__init__()
        self._module = []
        for i in range(num_layers-1):
            if i==0:
                self._module.append(nn.Linear(in_dim,hidden_dim)) 
            else:
                self._module.append(nn.Linear(hidden_dim,hidden_dim))
            self._module.append(nn.ReLU())
        self._module.append(nn.Linear(hidden_dim,1))
        self._module.append(nn.BatchNorm1d(1))
        self._module.append(nn.Sigmoid())
        self._module = nn.Sequential(*self._module)
    def forward(self, hidden, candidate_mask, z, no_new):
        if self.training:
            z = z/(z.max()+1e-10)
        hidden = hidden.view([-1,hidden.shape[-1]])
        p = self._module(hidden)
        p = p.view([candidate_mask.shape[0],-1,p.shape[-1]])
        
        t = torch.sum(candidate_mask*(1.0-z),1,keepdims=True)
        t = (t>=no_new-0.5).type(torch.FloatTensor).detach().to(p.device)

        p = p*(1-t+t*candidate_mask)*(1.0-z)
        #p = p*candidate_mask*(1.0-t*z)
        #p = 1e-10 + (1-2e-10)*p
        p = p / (torch.sum(p,1,keepdims=True)+1e-10)
        p = p.squeeze(2)
        if self.training:
            temp = torch.Tensor([0.05]).to(hidden.device)
            p_z = RelaxedBernoulli(temp, probs=p*no_new)
            z_i = p_z.rsample()
        else:
            z_i = p.new_zeros(p.shape)
            for i in range(p.shape[0]):
                try:
                    pick = np.random.choice(p.shape[1],no_new,replace=False,p=p[i].squeeze().to('cpu').data.numpy())
                except Exception:
                    pdb.set_trace()
                z_i[i,pick] = 1.0
            #z_i = OneHotCategorical(probs=p).sample()

        z_i = z_i.unsqueeze(2)

        return z_i

class SubsetSelect(nn.Module):
    def __init__(self, num_layers, point_dim, hidden_dim):
        super(SubsetSelect, self).__init__()
        self._module = []
        for i in range(num_layers-1):
            if i==0:
                in_dim = point_dim
            else:
                in_dim = hidden_dim
            self._module.append(nn.Linear(in_dim,hidden_dim))
            self._module.append(nn.ReLU())
        self._module.append(nn.Linear(hidden_dim,hidden_dim))
        self._module = nn.Sequential(*self._module)
    
        self._module2 = []
        for i in range(num_layers-1):
            if i==0:
                in_dim = point_dim
            else:
                in_dim = hidden_dim
            self._module2.append(nn.Linear(in_dim,hidden_dim))
            self._module2.append(nn.ReLU())
        self._module2.append(nn.Linear(hidden_dim,hidden_dim))
        self._module2 = nn.Sequential(*self._module2)
        self.dbb = BerDrop(1)

        self._module_dict = nn.ModuleDict()
        for i in range(1):
            if i == 0:
                Q_dim = hidden_dim
                K_dim = hidden_dim
            else:
                Q_dim = hidden_dim
                K_dim = hidden_dim
            self._module_dict['MAB'+str(i)] = MAB(Q_dim,K_dim,hidden_dim,4,True)
        self.num_layers = num_layers 
        self.autoreg = AutoRegressive(num_layers,hidden_dim,hidden_dim)
    def forward(self,points,k):
        hidden_set = self._module(points)
        hidden = self._module2(points)
        
        ### Doing candidate selection
        set_representation = torch.mean(hidden_set,1,keepdims=True)
        set_representation = set_representation.repeat(1,points.shape[1],1)

        mask, reg = self.dbb(torch.cat([hidden,set_representation],2),get_reg=True)
        # if not self.training:
        #     masked_points = [torch.masked_select(points[i],mask[i].ge(0.5)).view(1,-1,points.shape[2]) for i in range(len(points))]
        #     max_len = max([masked_point.shape[1] for masked_point in masked_points])
        #     mask = [torch.cat([p.new_ones([1,p.shape[1],1]),p.new_zeros([1,max_len-p.shape[1],1])],1) for p in masked_points]
        #     mask = torch.cat(mask,0)
        #     masked_points = [torch.cat([p,p.new_zeros([p.shape[0],max_len-p.shape[1],p.shape[2]])],1) for p in masked_points]
        #     points = torch.cat(masked_points,0)
        #     hidden = self._module2(points)

        ### Doing subset selection
        #set_representation = torch.sum(hidden_set*mask,1,keepdims=True) / (torch.sum(mask,1,keepdims=True)+1e-10)
        #set_representation = set_representation.repeat(1,points.shape[1],1)

        new_pick = 50
        if self.training:
            i = int((k+new_pick-1)/new_pick)*new_pick-new_pick
            #if i==0:
            #    i = 1e-2
            if i!=0:
                scale = mask.sum(1,keepdims=True)/(i+1e-10)
                temp = torch.Tensor([0.05]).to(mask.device)
                z = RelaxedBernoulli(temp, probs=mask/scale).rsample()
                z_real = Bernoulli(probs=mask/scale).sample()
            else:
                z = mask.new_zeros(mask.shape)
                z_real = mask.new_zeros(mask.shape)
            #h = torch.cat([hidden,set_representation],2)
            h = hidden
            for j in range(1):
                h = self._module_dict['MAB'+str(j)](h,hidden,z)
            z_i = self.autoreg(h,mask,z,no_new=new_pick)
            z = z + z_i

        else:
            z = torch.zeros_like(mask).to(mask.device)
            z_real = torch.zeros_like(mask).to(mask.device)
            for i in range(int((k+new_pick-1)/new_pick)):
                #h = torch.cat([hidden,set_representation],2)
                h = hidden
                for j in range(1):
                    h = self._module_dict['MAB'+str(j)](h,hidden,z)

                z_i = self.autoreg(h,mask,z,no_new=new_pick)
                z = z + z_i

            #points = torch.masked_select(points,z.ge(0.5)).view(points.shape[0],-1,points.shape[2])
            #z = points.new_ones([points.shape[0],points.shape[1],1])
            #mask = z

        return points, mask, reg, z, z_real




#### Stochastic Subset Selection: Combine of Decoder and Encoder 
class Model(nn.Module):
    def __init__(self, config):
        super(Model, self).__init__()

        self.train_with_real_mask = config.train_with_real_mask

        """StochasticSubset's Decoder, which is a CNP"""
        self._decoder = DeterministicANP(config.CNP_encoder_num_layers, config.CNP_decoder_num_layers, config.x_dim, config.y_dim, config.hidden_dim)

        """StochasticSubset's Encoder layers"""
        self._subsetSelect = SubsetSelect(config.subset_encoder_num_layers, config.x_dim+config.y_dim, config.hidden_dim)

        self.classify_model = Classification_Model() 

    def forward(self, x, y, index, attr, subset_size):
        if self.training:
            k = np.random.choice(subset_size)+1
        else:
            k = subset_size
        points = torch.cat([x,y],2)
        subset_points, candidate_mask, reg, subset_mask, subset_mask_real = self._subsetSelect(points,k)
        temp = torch.split(subset_points, x.shape[2], 2)
        subset_x = temp[0]
        subset_y = torch.cat(temp[1:],2)

        index = (subset_x*torch.Tensor([[[218*178,178]]]).to(subset_x.device)).sum(-1).round().to(torch.long).unsqueeze(-1).repeat([1,1,subset_y.shape[-1]])
        #index = index.unsqueeze(-1).repeat([1,1,subset_y.shape[-1]])
        img = subset_y.new_zeros([subset_y.shape[0],218*178,subset_y.shape[2]])
        pdb.set_trace()
        img.scatter_(1,index,subset_y*subset_mask)
        img = img.view(-1,218,178,img.shape[-1]).permute(0,3,1,2).contiguous()
        classify_loss, pred = self.classify_model(img,attr)

        classify_loss_greedy = None
        pred_greedy = None
        if self.training:
            with torch.no_grad():
                self._subsetSelect.eval()
                subset_points_greedy, candidate_mask_greedy, reg_greedy, subset_mask_greedy, subset_mask_real_greedy = self._subsetSelect(points,k)
                temp = torch.split(subset_points_greedy, x.shape[2], 2)
                subset_x_greedy = temp[0]
                subset_y_greedy = torch.cat(temp[1:],2)
                self._subsetSelect.train()
            index_greedy = (subset_x_greedy*torch.Tensor([[[218*178,178]]]).to(subset_x_greedy.device)).sum(-1).round().to(torch.long).unsqueeze(-1).repeat([1,1,subset_y_greedy.shape[-1]])
            img_greedy = subset_y_greedy.new_zeros([subset_y_greedy.shape[0],218*178,subset_y_greedy.shape[2]])
            img_greedy.scatter_(1,index_greedy,subset_y_greedy*subset_mask_greedy)
            img_greedy = img_greedy.view(-1,218,178,img_greedy.shape[-1]).permute(0,3,1,2).contiguous()
            classify_loss_greedy, pred_greedy = self.classify_model(img_greedy,attr)


        return classify_loss, pred, classify_loss_greedy, pred_greedy, subset_x, subset_y, subset_mask, candidate_mask, reg


class DataParallel_Model(nn.Module):
    def __init__(self, config):
        super(DataParallel_Model, self).__init__()
        self.train_with_real_mask = config.train_with_real_mask

        self._model = Model(config)
        self._model = nn.DataParallel(self._model)

    def forward(self, x, y, index, attr, subset_size=100):
        classify_loss, pred, classify_loss_greedy, pred_greedy, subset_x, subset_y, subset_mask, candidate_mask, reg = self._model(x, y, index, attr, subset_size)

        loss = 0
        classify_loss = classify_loss.sum(1).mean()
        loss += classify_loss
        if classify_loss_greedy is not None:
            classify_loss_greedy = classify_loss_greedy.sum(1).mean()
            loss += classify_loss_greedy
        if reg is not None:
            reg = reg.sum(1).mean()
            loss += reg
        pred_dict = OrderedDict()
        pred_dict['Main'] = pred.to('cpu').data
        if pred_greedy is not None:
            pred_dict['Greedy'] = pred_greedy.to('cpu').data

        train_dict = OrderedDict()
        train_dict['Classification loss'] = classify_loss.item()
        if classify_loss_greedy is not None:
            train_dict['Classification loss greedy'] = classify_loss_greedy.item()
        if reg is not None:
            train_dict['Reg'] = reg.to('cpu').item()
        if subset_mask is not None:
            train_dict['Subset Mask usage'] = (subset_mask>0.5).sum().item()/np.prod(subset_mask.shape)
        if candidate_mask is not None:
            train_dict['Candidate Mask usage'] = candidate_mask.sum().item()/np.prod(candidate_mask.shape)
        

        test_dict = OrderedDict()
        test_dict['subset_x'] = subset_x
        test_dict['subset_y'] = subset_y
        test_dict['subset_mask'] = subset_mask
        test_dict['candidate_mask'] = candidate_mask

        return loss, pred_dict, train_dict, test_dict


