from __future__ import absolute_import, division, print_function, unicode_literals
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import numpy as np
#import matplotlib as mpl
#mpl.use('Agg')
#import matplotlib.pyplot as plt
import collections
from tqdm import tqdm
import pdb
from datetime import datetime
from collections import OrderedDict
import importlib
from collections import defaultdict
from sklearn.metrics import roc_auc_score

DATASET = 'celeba'          
MODEL_NAME = 'classify_selected_points' 
MODE = 'training'                   # training or inference

CHECKPOINT = 'saved/classify_selected_points_with_rec_celeba9.pth'

NUM_EPOCHS = 100
BATCH_SIZE = 64
VALIDATE_WHILE_TRAINING = True


# Model config
config = lambda: None


config.hidden_dim = 128
config.CNP_encoder_num_layers = 4
config.CNP_decoder_num_layers = 2

# Used for variational subset 
config.subset_encoder_num_layers = 3

# Use for learning to sample
config.sampler_num_output_points = 100
config.sampler_num_conv_layers = 4
config.sampler_num_fc_layers = 4

config.train_with_real_mask = False #only for stochastic_subset

from celeba import *
train_set = CELEBA(train=True)
train_loader = DataLoader(train_set, batch_size=BATCH_SIZE, shuffle=True, num_workers=16)

test_set = CELEBA(train=False)
test_loader = DataLoader(test_set, batch_size=BATCH_SIZE, shuffle=False, num_workers=16)
config.x_dim = 2
config.y_dim = 3
VALIDATE_WHILE_TRAINING = True
config.max_output_points = 500
config.num_points_jump = 100



device = 'cuda'

# Define the model
model_class = getattr(importlib.import_module('models.'+MODEL_NAME), 'DataParallel_Model')
model = model_class(config)
model.to(device)



model_name = '%s_%s'%(MODEL_NAME,DATASET)

# Set up the optimizer and train step
optimizer = optim.Adam(model.parameters(),1e-3)
lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.96)
if MODE == 'training':

    def train_step(x, y, index, attr):
        model.train()

        x = x.to(device)
        y = y.to(device)
        index = index.to(device)
        attr = attr.to(device).to(torch.float)

        optimizer.zero_grad()

        loss, pred_dict, train_dict, _ = model(x,y,index,attr)
        loss.backward()
        optimizer.step()
        del loss

        return pred_dict, train_dict



    def test_step(x, y, index, attr):
        model.eval()

        x = x.to(device)
        y = y.to(device)
        index = index.to(device)
        attr = attr.to(device).to(torch.float)
        _, pred_dict, train_dict, test_dict = model(x, y, index, attr)

        return pred_dict, train_dict, test_dict




    for epoch in range(NUM_EPOCHS):
        print('\nEpoch %d:'% (epoch), model_name)
        print('\t Training:')
        epoch_train_dict = defaultdict(list)
        epoch_pred_dict = defaultdict(list)
        epoch_attr = []
        for it, (x, y, index, attr) in tqdm(enumerate(train_loader),total=len(train_loader),ncols=75,leave=False):
            pred_dict, train_dict = train_step(x,y,index,attr)
            for key in train_dict:
                epoch_train_dict[key].append(train_dict[key])
            for key in pred_dict.keys():
                epoch_pred_dict[key].append(pred_dict[key])
            epoch_attr.append(attr)


        for key in epoch_train_dict.keys():
            print('\t%s: %.4f'%(key,np.mean(epoch_train_dict[key])))
        epoch_attr = np.concatenate(epoch_attr,0)
        for key in epoch_pred_dict.keys():
            pred = np.concatenate(epoch_pred_dict[key],0)
            auc = roc_auc_score(epoch_attr,pred,average=None)
            print('Auc %s: '%(key), auc)

        torch.save(model.state_dict(), 'saved/'+model_name+str(epoch+1)+'.pth')
        if VALIDATE_WHILE_TRAINING:
            with torch.no_grad():
                print('\t Validate:')
                epoch_train_dict = defaultdict(list)
                epoch_pred_dict = defaultdict(list)
                epoch_attr = []
                for it, (x, y, index, attr) in tqdm(enumerate(test_loader),total=len(test_loader),ncols=75,leave=False):
                    pred_dict, train_dict,_ = test_step(x,y,index,attr)
                    for key in train_dict:
                        epoch_train_dict[key].append(train_dict[key])
                    for key in pred_dict.keys():
                        epoch_pred_dict[key].append(pred_dict[key])
                    epoch_attr.append(attr)


                for key in epoch_train_dict.keys():
                    print('\t%s: %.4f'%(key,np.mean(epoch_train_dict[key])))
                epoch_attr = np.concatenate(epoch_attr,0)
                for key in epoch_pred_dict.keys():
                    pred = np.concatenate(epoch_pred_dict[key],0)
                    auc = roc_auc_score(epoch_attr,pred,average=None)
                    print('Auc %s: '%(key), auc)

        lr_scheduler.step()

elif MODE == 'inference':
    model_dict = torch.load(CHECKPOINT)
    model.load_state_dict(model_dict)


    def test_step(x, y, index, attr):
        model.eval()

        x = x.to(device)
        y = y.to(device)
        index = index.to(device)
        attr = attr.to(device).to(torch.float)
        _, pred_dict, train_dict, test_dict = model(x, y, index, attr)

        return pred_dict, train_dict, test_dict
    

    with torch.no_grad():
        print('\t Inference:')
        epoch_train_dict = defaultdict(list)
        epoch_pred_dict = defaultdict(list)
        epoch_attr = []
        for it, (x, y, index, attr) in tqdm(enumerate(test_loader),total=len(test_loader),ncols=75,leave=False):
            pred_dict, train_dict,_ = test_step(x,y,index,attr)
            for key in train_dict:
                epoch_train_dict[key].append(train_dict[key])
            for key in pred_dict.keys():
                epoch_pred_dict[key].append(pred_dict[key])
            epoch_attr.append(attr)


        for key in epoch_train_dict.keys():
            print('\t%s: %.4f'%(key,np.mean(epoch_train_dict[key])))
        epoch_attr = np.concatenate(epoch_attr,0)
        for key in epoch_pred_dict.keys():
            pred = np.concatenate(epoch_pred_dict[key],0)
            auc = roc_auc_score(epoch_attr,pred,average=None)
            print('Auc %s: %.5f'%(key,np.mean(auc)), auc)

