import torch
import torch.nn as nn
import pdb
import torch.nn.functional as F
from torch.distributions.bernoulli import Bernoulli
import numpy as np
from collections import OrderedDict
from models.anp import *



class RandomSelect(nn.Module):
    def __init__(self, num_layers, x_dim, y_dim, hidden_dim):
        super(RandomSelect, self).__init__()
        
    def forward(self,x,y,i):
        idx = torch.from_numpy(np.random.choice(x.shape[1],i)).to(x.device)
        subset_x = torch.index_select(x,1,idx)
        subset_y = torch.index_select(y,1,idx)

        return subset_x, subset_y, torch.ones(subset_x.shape[0],subset_x.shape[1],1).to(x.device)






class Model(nn.Module):
    def __init__(self, config):
        super(Model, self).__init__()

        """Decoder, which is a CNP"""
        self._decoder = DeterministicANP(config.CNP_encoder_num_layers, config.CNP_decoder_num_layers, config.x_dim, config.y_dim, config.hidden_dim)

        """Random Selection"""
        self._encoder = RandomSelect(config.subset_encoder_num_layers, config.x_dim, config.y_dim, config.hidden_dim)

    def forward(self, x, y, i):
        subset_x, subset_y, mask = self._encoder(x,y,i)
        log_p , mu, var = self._decoder(subset_x, subset_y , x, y, mask)

        return log_p, mu, var, subset_x, subset_y

class DataParallel_Model(nn.Module):
    def __init__(self, config):
        super(DataParallel_Model, self).__init__()
        self.model = Model(config)
        self.model = nn.DataParallel(self.model)
        self.max_output_points = config.max_output_points
        self.num_points_jump = config.num_points_jump

    def forward(self, x, y, k=100):
        if self.training:
            i = int(np.random.choice(self.max_output_points)/self.num_points_jump)*self.num_points_jump+self.num_points_jump
        else:
            i = k
        log_p, mu, var, subset_x, subset_y = self.model(x,y,i)

        loss_p = -log_p.sum(1).mean()
        loss = loss_p

        train_dict = OrderedDict()
        train_dict['Reconstruction loss'] = loss_p.to('cpu').item()

        test_dict = OrderedDict()
        test_dict['subset_x'] = subset_x
        test_dict['subset_y'] = subset_y
        test_dict['mu'] = mu 
        test_dict['var'] = var

        return loss, train_dict, test_dict

